import aiohttp
from ._d4 import *
import typing
import asyncio
from ._utils import *
import time
import sys
import traceback
from ._exception import *
import json
from ._cmd import *
from ._models import *

v9_url = 'https://discord.com/api/v9'


def Log(message=None):
    if printLog:
        print(message)
    else:
        pass


class AsyncHTTPRequest:
    def __init__(self,
                 token: str,
                 loop: asyncio.AbstractEventLoop = None,
                 session: aiohttp.ClientSession = None,
                 retry_default=3):
        self.loop = loop or asyncio.get_event_loop()
        self.token = token
        self.session = session or aiohttp.ClientSession(loop=self.loop)
        self.retry_default = retry_default
        self._closed = False

    async def close(self):
        await self.session.close()
        self._closed = True

    async def request(self, route: str, meth: str, body: typing.Any = None, *, is_json: bool = False, reason_header: str = None, retry: int = None, **kwargs) -> dict:
        code = 429  # Empty code in case of rate limit fail.
        resp = {}   # Empty resp in case of rate limit fail.
        retry = (retry if retry >
                 0 else 1) if retry is not None else self.default_retry
        for x in range(retry):
            code, resp = await self._request(route, meth, body, is_json, reason_header, **kwargs)
            if 200 <= code < 300:
                return resp
            elif code == 400:
                raise BadRequest(route, code, resp)
            elif code == 403:
                raise Forbidden(route, code, resp)
            elif code == 404:
                raise NotFound(route, code, resp)
            elif code == 429:
                wait_sec = resp["retry_after"]
                await asyncio.sleep(wait_sec)
                continue
            elif 500 <= code < 600:
                raise DiscordError(code)
            else:
                raise Unknown(code)
        raise RateLimited(route, code, resp)

    async def _request(self, route: str, meth: str, body: typing.Any = None, is_json: bool = False, reason_header: str = None, **kwargs) -> typing.Tuple[int, dict]:
        headers = {"Authorization": f"Bot {self.token}"}
        if meth not in ["GET"] and body is not None:
            if is_json:
                headers["Content-Type"] = "application/json"
                body = json.dumps(body)
            kwargs["data"] = body
        if reason_header is not None:
            headers["X-Audit-Log-Reason"] = reason_header
        async with self.session.request(meth, self.BASE_URL+route, headers=headers, **kwargs) as resp:
            return resp.status, await resp.json() if resp.status != 204 else None


class WSClosing(Exception):
    # def __init__(self, code):
    #     self.code = code
    pass


async def safe_call(coro, additional_message: typing.Optional[str] = None):
    """
    Calls coroutine, ignoring raised exception and only print traceback.
    This is used for event listener call, and intended to be used at creating task.
    :param coro: Coroutine to safely call
    :param additional_message: Additional traceback message to print at the top.
    """

    try:
        await coro
    except Exception as ex:
        tb = traceback.format_exc()
        if additional_message:
            _p = additional_message + "\n" + tb
        else:
            _p = tb
        print(_p, file=sys.stderr)


def dispatch(self, name, *args):
    """
    Dispatches new event.
    :param name: Name of the event.
    :param args: Arguments of the event.
    """
    [self.loop.create_task(safe_call(x(*args)))
     for x in self.events.get(name.upper())]
    # [self.__wait_futures[name.upper()].pop(x).set_result(args) for x in range(len(self.__wait_futures.get(name.upper(), [])))]
    for x in range(len(self.__wait_futures.get(name.upper(), []))):
        fut: asyncio.Future = self.__wait_futures[name.upper()].pop(x)
        if not fut.cancelled():
            fut.set_result(args)


def dispatch_from_raw(self, name, resp):
    ret = self.process_response(name, resp)
    if hasattr(ret, "_dont_dispatch") and ret._dont_dispatch:
        return
    self.client.dispatch(name, ret)


class WebSocketClient:
    def __init__(self, token, intents: typing.Union[Intents, int] = None, ws: aiohttp.ClientWebSocketResponse = None,):
        self.http = AsyncHTTPRequest
        self.ws = ws or aiohttp.ClientWebSocketResponse
        self.base_url: str = ""
        self.intents = intents or typing.Union[Intents, int]
        self.session = aiohttp.ClientSession
        self.heartbeat_interval: int
        self._closed = False
        self.last_heartbeat_send = 0
        self.last_heartbeat_ack = 0
        self._reconnecting = False
        self._fresh_reconnecting = False
        self._heartbeat_task = None
        self.session_id = None
        self.seq = None
        self._ping_start = 0.0
        self.token = token
        self.ping = 0.0

    async def close(self, code: int = 1000):
        if self._closed:
            return
        Log(self.ws)
        await self.ws.close(code=code)
        self._closed = True

    async def reconnect(self, fresh: bool = False):
        if self._reconnecting or self._fresh_reconnecting:
            return
        self._reconnecting = not fresh
        self._fresh_reconnecting = fresh
        if not self._closed:
            await self.close(4000)

    async def identify(self):
        data = {
            "op": Opcodes.IDENTIFY,
            "d": {
                "token": self.token,
                "intents": int(self.intents),
                "properties": {
                    "$os": sys.platform,
                    "$browser": "d4-buttons-dico",
                    "$device": "d4-buttons-dico"
                }
            }
        }
        await self.ws.send_json(data)

    async def resume(self):
        data = {
            "op": Opcodes.RESUME,
            "d": {
                "token": self.token,
                "session_id": self.session_id,
                "seq": self.seq
            }
        }

    async def send_heartbeat(self):
        while not self._closed:
            if not self.last_heartbeat_send <= self.last_heartbeat_ack <= time.time():
                await self.reconnect()
                break
            data = {"op": Opcodes.HEARTBEAT, "d": self.seq}
            self._ping_start = time.time()
            await self.ws.send_json(data)
            self.last_heartbeat_send = time.time()
            await asyncio.sleep(self.heartbeat_interval/1000)

    async def connect(self):
        while True:
            while not self._closed:
                try:
                    ws = await self.receive()
                    res = ws.json()
                    Log(res)
                except WSClosing as ex:
                    # res = await ws.receive()
                    # res = res.json()
                    # print(res)
                    # if res["s"]:
                    #     self.seq = res["s"]

                    # if res["op"] == Opcodes.HELLO:
                    #     print("OPCODE HELLO RECEIVED")
                    #     self.ws = res
                    #     self.heartbeat_interval = res["d"]["heartbeat_interval"]
                    #     self._heartbeat_task = asyncio.get_event_loop().create_task(
                    #         self.send_heartbeat(ws, self.heartbeat_interval))
                    #     if self._reconnecting:
                    #         await self.resume()
                    #     else:
                    #         await self.identify()
                    #     if self._fresh_reconnecting:
                    #         self._fresh_reconnecting = False
                    # elif res["op"] == Opcodes.DISPATCH:
                    #     if res["t"] == "READY":
                    #         self.session_id = res["d"]["session_id"]
                    #         # dispatch
                    # elif res["op"] == Opcodes.HEARTBEAT_ACK:
                    #     self.last_heartbeat_ack = time.time()
                    #     self.ping = self.last_heartbeat_ack - self._ping_start

                    # elif res["op"] == Opcodes.INVALID_SESSION:
                    #     await asyncio.sleep(5)
                    #     await self.reconnect(fresh=True)
                    #     break

                    # elif res["op"] == Opcodes.RECONNECT:
                    #     await self.reconnect()
                    #     break
                    Log("Websocket is closing with code:" + {ex.code})
                    if ex.code is None and self.try_reconnect:
                        Log("Trying to reconnect...")
                        await self.reconnect(fresh=True)
                    break
                a = Opcodes.as_string(res["op"])
                b = ' with event name `' + (res["t"] if res["t"] else "None")
                c = res["op"]
                d = res["t"]
                e = f"Received {a} payload" f"{f' with event name {d}' if res['op'] == Opcodes.DISPATCH else ''}."
                Log(e)

                if res["s"]:
                    self.seq = res["s"]

                if res["op"] == Opcodes.DISPATCH:
                    global event_dict
                    global command_dict
                    if res["t"] == "READY":
                        self.session_id = res["d"].get(
                            "session_id", self.session_id)
                        if "ON_READY".lower() in event_dict.keys():
                            func = event_dict["on_ready"]
                            await func()
                    elif res["t"] == "INTERACTION_CREATE":
                        if "INTERACTION_CREATE".lower() in event_dict.keys():
                            func = event_dict["interaction_create"]
                            e = Interaction(**res["d"])
                            await func(e)  # TODO: call function with info
                    elif res["t"] == "MESSAGE_CREATE":
                        if "MESSAGE_CREATE".lower() in event_dict.keys():
                            func = event_dict["message_create"]
                            e = Message(**res["d"])
                            await func(e)
                        else:
                            e = Message(**res["d"])
                            for i in command_dict.keys():
                                if str(e.content) == i:
                                    func = command_dict[e.content]
                                    await func(e)
                                elif str(e.content).startswith(i + ' '):
                                    func = command_dict[i]
                                    await func(e)
                    else:
                        if res["t"].lower() in event_dict.keys():
                            func = event_dict[res["t"]]
                            await func(**res["d"])

                elif res["op"] == Opcodes.HELLO:
                    self.heartbeat_interval = res["d"]["heartbeat_interval"]
                    self._heartbeat_task = asyncio.get_event_loop().create_task(self.send_heartbeat())
                    if self._reconnecting:
                        await self.resume()
                    else:
                        await self.identify()
                    if self._fresh_reconnecting:
                        self._fresh_reconnecting = False

                elif res["op"] == Opcodes.HEARTBEAT_ACK:
                    self.last_heartbeat_ack = time.time()
                    self.ping = self.last_heartbeat_ack - self._ping_start

                elif res["op"] == Opcodes.INVALID_SESSION:
                    print(
                        "Failed utils resume, reconnecting to utils without resuming.")
                    await asyncio.sleep(5)
                    await self.reconnect(fresh=True)
                    break

                elif res["op"] == Opcodes.RECONNECT:
                    await self.reconnect()
                    break

            if self._reconnecting or self._fresh_reconnecting:
                self._closed = False
            else:
                return

    async def receive(self):
        res = await self.ws.receive()
        if res.type == aiohttp.WSMsgType.TEXT:
            self.seq = res.json()["s"]
            return res
        elif res.type in (aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSING, aiohttp.WSMsgType.CLOSED):
            raise WSClosing(res.data)

    @classmethod
    async def raiseConnection(cls, tk, it):
        ws = await aiohttp.ClientSession().ws_connect(url="wss://gateway.discord.gg/?v=9&encoding=json")
        return cls(tk, intents=it, ws=ws)



def Buttons(*_buttons):
    async def sendAll(ctx=None, content=None, id=None):
        i = 0

        r = []
        for button in _buttons:
            i += 1
            if i > 5:
                raise SyntaxError.msg == "You can only request 5 buttons at a time."
            self = button
            headers = {"Authorization": f"Bot {Token}"}
            headers["Content-Type"] = "application/json"
            channel_id: str
            if ctx is None and id is None:
                raise TypeError.msg == "Error: Code 10. Check out https://github.com/namuKR/d4-buttons/wiki#code-10 for more info."
            elif ctx is not None and id is not None:
                channel_id = ctx.channel.id
            elif ctx is None and id is not None:
                channel_id = id
            elif ctx is not None and id is None:
                channel_id = ctx.channel.id

            if content is None:
                print(
                    "Warning: Code 10. Check out https://github.com/namuKR/d4-buttons/wiki#code-10-1 for more info.")
                content = '​'  # zero width space inside it!

            if self.url is not None and self.customId is None and self.emoji is None:  # Link Button

                r[i] = {
                    "type": 2,
                    "label": self.text,
                    "style": 5,
                    "url": self.url,
                    "disabled": self.disabled if self.disabled else False
                }

            elif self.url is not None and self.customId is None and self.emoji is not None:

                r[i] = {
                    "type": 2,
                    "label": self.text,
                    "style": 5,
                    "url": self.url,
                    "disabled": self.disabled if self.disabled else False,
                    "emoji": self.emoji
                }

            elif self.emoji is not None:
                if self.customId is None:
                    raise TypeError.msg == "Buttons that are not Link style must have customId values."
                elif self.style is None or self.text is None:
                    raise TypeError.msg == "Button.style or Button.text missing"

                r[i] = {
                    "type": 2,
                    "label": self.text,
                    "style": self.style,
                    "emoji": self.emoji,
                    "custom_id": self.customId,
                    "disabled": self.disabled if self.disabled else False
                }

            elif self.url is not None and self.customId is not None:
                raise TypeError.msg == "Link Buttons should not have customId values."
            elif self.url is None and self.customId is None:
                raise TypeError.msg == "Buttons that are not Link style must have customId values."
            elif self.style is None or self.text is None:
                raise TypeError.msg == "Button.style or Button.text missing"
            elif type(self.style) is not int:
                raise TypeError.msg == "Button.style is not int - button style must be int type."
            else:

                r[i] = {
                    "type": 2,
                    "label": self.text,
                    "style": self.style,
                    "custom_id": self.customId,
                    "disabled": self.disabled if self.disabled else False
                }

        rq: list
        for v in range(i):
            rq.append(r[v])

        req_body: dict = {
            "content": content,
            "tts": False,
            "components": [
                {
                    "type": 1,
                    "components": [

                    ]
                }
            ]

        }

        req_body["components"][0]["components"] = rq

        response = requests.post(
            v9_url + f"/channels/{channel_id}/messages", headers=headers, json=req_body)
        return response


"""
d4.py
~~~~~
Another Discord API wrapper.
Supports v9.

:copyright: (c) 2021-present ast.08
:license: Apache License 2.0, see LICENSE for more details.
"""
__title__ = 'd4.py'
__author__ = 'ast.08'
__license__ = 'Apache License 2.0'
__copyright__ = 'Copyright 2021-present ast.08'
__version__ = '0.0.0.10a2'

from ._d4 import *
from ._interactions import *
from ._exception import *
from ._utils import *
from ._cmd import *
from ._requests import *
from ._models import *



from ._d4 import *
import typing
import inspect
from ._requests import *
from ._models import *


def nitro_converter(value):
    return {
        0 or '0': None,
        1 or '1': 'Nitro Classic',
        2 or '2': 'Nitro'
    }[value]


def visibility_converter(value):
    return {
        0 or '0': None,
        1 or '1': 'Everyone'
    }[value]


def get_current_user():
    return User(**get('/users/@me').json())


def get_user(id_):  # TODO: Check if ** works or not
    id_ = str(id_)
    return User(**get(f'users/{id_}').json())


def modify_current_user(username=None, avatar=None):
    return User(**patch('users/@me', {'username': username, 'avatar': avatar}).json())


def get_current_user_guilds():  # Change to Guild() when finished
    return get('/users/@me/guilds').json()


def leave_guild(id_):
    return delete(f'/users/@me/guilds/{id_}').json()


def create_dm(id_):
    dm_channel = post(
        f'/users/@me/channels', {'recipient_id': id_}).json()
    return dm_channel


def create_group_dm(id_, acc_token: list, nicks: dict):
    group_channel = post(
        f'/users/@me/channels', {'access_tokens': acc_token, 'nicks': nicks})
    return group_channel


def get_user_connections():
    return get('/users/@me/connections')


def activity_type_converter(act):
    if isinstance(act, Activity):
        return {
            0: f"Playing {act.name}",
            1: f"Streaming {act.details}",
            2: f"Listening to {act.name}",
            3: f"Watching {act.name}",
            4: f"{act.emoji} {act.name}",
            5: f"Competing in {act.name}"
        }[act.type]
    else:
        return {
            0: "Game",
            1: "Streaming",
            2: "Listening",
            3: "Watching",
            4: "Custom",
            5: "Competing"
        }[act]


def get_channel(channel_id=None):
    return Channel(**get(f"/channels/{channel_id}").json())

# From https://github.com/eunwoo1104/dico - dico/model/gateway.py


class FlagBase:
    def __init__(self, *args, **kwargs):
        self.values = {x: getattr(self, x) for x in dir(
            self) if isinstance(getattr(self, x), int)}
        self.value = 0
        for x in args:
            if x.upper() not in self.values:
                raise AttributeError(f"invalid name: `{x}`")
            self.value |= self.values[x.upper()]
        for k, v in kwargs.items():
            if k.upper() not in self.values:
                raise AttributeError(f"invalid name: `{k}`")
            if v:
                self.value |= self.values[k.upper()]

    def __int__(self):
        return self.value

    def __getattr__(self, item):
        return self.has(item)

    def has(self, name: str):
        if name.upper() not in self.values:
            raise AttributeError(f"invalid name: `{name}`")
        return (self.value & self.values[name.upper()]) == self.values[name.upper()]

    def __setattr__(self, key, value):
        orig = key
        key = key.upper()
        if orig in ["value", "values"] or key not in self.values.keys():
            return super().__setattr__(orig, value)
        if not isinstance(value, bool):
            raise TypeError(f"only type `bool` is supported.")
        has_value = self.has(key)
        if value and not has_value:
            self.value |= self.values[key]
        elif not value and has_value:
            self.value &= ~self.values[key]

    def add(self, value):
        return self.__setattr(value, True)

    def remove(self, value):
        return self.__setattr(value, False)

    @classmethod
    def from_value(cls, value: int):
        ret = cls()
        ret.value = value
        return ret


class Intents(FlagBase):
    GUILDS = 1 << 0
    GUILD_MEMBERS = 1 << 1
    GUILD_BANS = 1 << 2
    GUILD_EMOJIS = 1 << 3
    GUILD_INTEGRATIONS = 1 << 4
    GUILD_WEBHOOKS = 1 << 5
    GUILD_INVITES = 1 << 6
    GUILD_VOICE_STATES = 1 << 7
    GUILD_PRESENCES = 1 << 8
    GUILD_MESSAGES = 1 << 9
    GUILD_MESSAGE_REACTIONS = 1 << 10
    GUILD_MESSAGE_TYPING = 1 << 11
    DIRECT_MESSAGES = 1 << 12
    DIRECT_MESSAGE_REACTIONS = 1 << 13
    DIRECT_MESSAGE_TYPING = 1 << 14

    @classmethod
    def full(cls):
        return cls(*[x for x in dir(cls) if isinstance(getattr(cls, x), int)])

    @classmethod
    def no_privileged(cls):
        ret = cls.full()
        ret.guild_presences = False
        ret.guild_members = False
        return ret


class Opcodes:
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    PRESENCE_UPDATE = 3
    VOICE_STATE_UPDATE = 4
    RESUME = 6
    RECONNECT = 7
    REQUEST_GUILD_MEMBERS = 8
    INVALID_SESSION = 9
    HELLO = 10
    HEARTBEAT_ACK = 11

    @staticmethod
    def as_string(code: int):
        opcodes = {0: "Dispatch",
                   1: "Heartbeat",
                   2: "Identify",
                   3: "Presence Update",
                   4: "Voice State Update",
                   6: "Resume",
                   7: "Reconnect",
                   8: "Request Guild Members",
                   9: "Invalid Session",
                   10: "Hello",
                   11: "Heartbeat ACK"}
        return opcodes.get(code)


def ensure_coro(func):
    async def wrap(*args, **kwargs):
        if inspect.iscoroutinefunction(func):
            return await func(*args, **kwargs)
        else:
            return func(*args, **kwargs)
    return wrap


class SessionStartLimit:
    def __init__(self, resp: dict):
        self.total = resp["total"]
        self.remaining = resp["remaining"]
        self.reset_after = resp["reset_after"]
        self.max_concurrency = resp["max_concurrency"]

    @classmethod
    def optional(cls, resp: dict):
        if resp:
            return cls(resp)

    def to_dict(self):
        return {"total": self.total, "remaining": self.remaining, "reset_after": self.reset_after, "max_concurrency": self.max_concurrency}


class GetGateway:
    def __init__(self, resp: dict):
        self.url: str = resp["url"]
        self.shards: typing.Optional[int] = resp.get("shards", 0)
        self.session_start_limit: typing.Optional[SessionStartLimit] = SessionStartLimit.optional(
            resp.get("session_start_limit"))

    def to_dict(self):
        return {"url": self.url, "shards": self.shards, "session_start_limit": self.session_start_limit}


class GatewayResponse:
    def __init__(self, resp: dict):
        self.raw = resp
        self.op = resp["op"]
        self.d = resp.get("d", {})
        self.s = resp.get("s")
        self.t = resp.get("t")

    def to_dict(self):
        return {"op": self.op, "d": self.d, "s": self.s, "t": self.t}


class EventHandler:
    def __init__(self, client):
        self.events = {}
        self.client = client

    def add(self, event, func):
        if event not in self.events:
            self.events[event] = []

        self.events[event].append(ensure_coro(func))

    def get(self, event) -> list:
        return self.events.get(event, [])

    def process_response(self, name, resp):
        model_dict = {
            "INTERACTION_CREATE": Interaction
        }
        if name in model_dict:
            ret = model_dict[name].create(self.client, resp)
        else:
            ret = resp
        return ret

    def dispatch_from_raw(self, name, resp):
        ret = self.process_response(name, resp)
        if hasattr(ret, "_dont_dispatch") and ret._dont_dispatch:
            return
        self.client.dispatch(name, ret)



import __init__ as d4

client = d4.Client(prefix="!", print_log=False)


@client.EventListener
async def ready():
    print("Bot is ready!")


@client.EventListener
async def interaction_create(interaction):
    if interaction.data.custom_id:
        await interaction.reply(f"The custom id for this button is {interaction.data.custom_id}.")


@client.cmd(name="ping")
async def ping(msg):
    await msg.get_channel().send("pong!")


@client.cmd()
async def time(msg):
    import time
    await msg.channel.send(time.strftime("%Y-%m-%d"))


@client.cmd()
async def me(msg):
    await msg.channel.send(f"Name: {msg.author.username}\nID: {msg.author.id}")


@client.cmd(name="createButton", aliases=["button", "create", "buttons"])
async def buttons_(msg):
    myButton = d4.Button(type=1, text="Test Button", customId="simpleId")
    await myButton.send(msg)

client.login("ODYzNjc2NTIwODA2NDgxOTIx.YOqXcA.170WwW7xjC8UOUZ8KcxPp9N3_k" + "k")



import requests
from ._d4 import *
base_url = 'https://discord.com/api/v9'


def post(url, data):
    headers = {"Authorization": f"Bot {Token}"}
    headers["Content-Type"] = "application/json"

    response = requests.post(base_url + url, json=data, headers=headers)
    return response


def get(url, data=None):
    headers = {"Authorization": f"Bot {Token}"}
    headers["Content-Type"] = "application/json"

    response = requests.get(base_url + url, headers=headers)
    return response


def patch(url, data):
    headers = {"Authorization": f"Bot {Token}"}
    headers["Content-Type"] = "application/json"

    response = requests.patch(base_url + url, json=data, headers=headers)
    return response


def delete(url):

    headers = {"Authorization": f"Bot {Token}"}
    headers["Content-Type"] = "application/json"

    response = requests.delete(base_url + url, headers=headers)
    return response



def BadRequest(*_):
    print("400 Bad Request")
    raise ConnectionError


def Forbidden(*_):
    print("403 Forbidden")
    raise ConnectionError


def NotFound(*_):
    print("404 Not Found")
    raise ConnectionError


def DiscordError(code):
    print(f"{code} Error")
    raise ConnectionError


def Unknown(code):
    print(f"{code} Error")
    raise ConnectionError


def RateLimited(*_):
    print("Rate Limited")
    raise ConnectionError


class CommandRegistrationError(Exception):
    def error(*args, **kwargs):
        print(f"Command registration failed: Try another command name?")



"""d4-py"""
"""Copyright 2021 ast.08. Check LICENSE for more details."""


import requests
import asyncio
from ._interactions import *
from ._cmd import *
v9_url = 'https://discord.com/api/v9'
Token: str

global commandClass
printLog: bool


class Client:
    def __init__(self, prefix=None, print_log=False):
        global printLog
        self.prefix = prefix or "!"
        printLog = print_log

    def EventListener(self, func):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Event must be a coroutine function")

        def decorator(*args, **kwargs):
            if func.__name__ == 'ready':
                event_dict['on_ready'] = func
            elif func.__name__ == 'message':
                event_dict['message_create'] = func
            elif func.__name__ == "interaction":
                event_dict['interaction_create'] = func
            else:
                event_dict[func.__name__.lower()] = func

        return decorator()

    def toClsCmd(self, name=None, aliases=None, slash=None, prefix=None, cls=None):
        if cls is None:
            cls = Command

        def decorator(func):
            if isinstance(func, Command):
                raise TypeError("Cannot register command: 'Already a command'")
            return cls(func, name=name, aliases=aliases, slash=slash, prefix=prefix)

        return decorator

    def cmd(self, *_args, name=None, aliases: list = [], slash=None, prefix=None, **__kwargs):  # decorator arugments
        ____name____ = name
        ____aliases____ = aliases
        ____slash____ = slash
        ____prefix____ = prefix

        def decorator(func):
            if ____prefix____ is None:
                pf = "!"
            else:
                pf = ____prefix____

            if ____name____ is None:
                name_ = pf + func.__name__
            else:
                name_ = pf + ____name____
            if ____aliases____ is None:
                aliases_: list = [].append(name_)
            else:
                aliases_ = []
                for i in ____aliases____:
                    aliases_.append(pf + i)
            res_cmd = self.toClsCmd(name_, aliases_, ____slash____,
                                    pf)(func)

            return Command().add_command(res_cmd)
            # return func(*args, **kwargs)
        return decorator

    async def start(self, tk):
        intent: Intents = Intents.no_privileged()
        ws = await WebSocketClient(token=tk, intents=intent).raiseConnection(tk, intent)
        try:
            await ws.connect()
        except KeyboardInterrupt:
            print("Exiting...")
        # except Exception:
        #     print("UNEXPECTED EXCEPTION OCCURED")
        #     traceback.print_exc()
        finally:
            await ws.close()
            await ws.http(tk).close()

    def login(self, token: str = None):
        global Token
        if token is None:
            raise SyntaxError.msg == "token is required"
        else:
            Token = token
            asyncio.get_event_loop().run_until_complete(self.start(tk=Token))


def Emoji(name: str = None, id: str = None, animated: bool = False):
    return {'name': name, 'id': id, 'animated': str(animated).lower()}


class Button:
    def __init__(self, style, text, emoji=None, url=None, customId=None, disabled: bool = False):
        self.style = style
        self.text = text
        self.emoji = emoji
        self.url = url
        self.customId = customId
        self.disabled = disabled

    async def send(self, ctx=None, content=None, id=None):
        headers = {"Authorization": f"Bot {Token}"}
        headers["Content-Type"] = "application/json"
        channel_id: str
        if ctx is None and id is None:
            raise TypeError(
                "Error: Code 10. Check out https://github.com/namuKR/d4-buttons/wiki#code-10 for more info.")
        elif ctx is not None and id is not None:
            channel_id = ctx.channel.id
        elif ctx is None and id is not None:
            channel_id = id
        elif ctx is not None and id is None:
            channel_id = ctx.channel.id

        if content is None:
            print(
                "Warning: Code 10. Check out https://github.com/namuKR/d4-buttons/wiki#code-10-1 for more info.")
            content = '​'  # zero width space inside it!

        if self.url is not None and self.customId is None and self.emoji is None:
            req_body = {
                "content": content,
                "tts": False,
                "components": [
                    {
                        "type": 1,
                        "components": [
                            {
                                "type": 2,
                                "label": self.text,
                                "style": 5,
                                "url": self.url,
                                "disabled": self.disabled if self.disabled else False
                            }
                        ]
                    }
                ]
            }
        elif self.url is not None and self.customId is None and self.emoji is not None:
            req_body = {
                "content": content,
                "tts": False,
                "components": [
                    {
                        "type": 1,
                        "components": [
                            {
                                "type": 2,
                                "label": self.text,
                                "style": 5,
                                "url": self.url,
                                "disabled": self.disabled if self.disabled else False,
                                "emoji": self.emoji
                            }
                        ]
                    }
                ]
            }
        elif self.emoji is not None:
            if self.customId is None:
                raise TypeError.msg == "Buttons that are not Link style must have customId values."
            elif self.style is None or self.text is None:
                raise TypeError.msg == "Button.style or Button.text missing"

            req_body = {
                "content": content,
                "tts": False,
                "components": [
                    {
                        "type": 1,
                        "components": [
                            {
                                "type": 2,
                                "label": self.text,
                                "style": self.style,
                                "emoji": self.emoji,
                                "custom_id": self.customId,
                                "disabled": self.disabled if self.disabled else False
                            }
                        ]
                    }
                ]
            }

        elif self.url is not None and self.customId is not None:
            raise TypeError.msg == "Link Buttons should not have customId values."
        elif self.url is None and self.customId is None:
            raise TypeError.msg == "Buttons that are not Link style must have customId values."
        elif self.style is None or self.text is None:
            raise TypeError.msg == "Button.style or Button.text missing"
        elif type(self.style) is not int:
            raise TypeError.msg == "Button.style is not int - button style must be int type."
        else:
            req_body = {
                "content": content,
                "tts": False,
                "components": [
                    {
                        "type": 1,
                        "components": [
                            {
                                "type": 2,
                                "label": self.text,
                                "style": self.style,
                                "custom_id": self.customId,
                                "disabled": self.disabled if self.disabled else False
                            }
                        ]
                    }
                ]
            }
        response = requests.post(
            v9_url + f"/channels/{channel_id}/messages", headers=headers, json=req_body)
        return response


chn: str
cntt = '​'


def queue(ctx=None, content=None, id=None):
    global chn, cntt
    id = str(id)
    if ctx is None and id is None:
        raise TypeError.msg == "Error: Code 10. Check out https://github.com/namuKR/d4-buttons/wiki#code-10 for more info."
    elif ctx is not None and id is not None:
        chn = ctx.channel.id
    elif ctx is None and id is not None:
        chn = id
    elif ctx is not None and id is None:
        chn = ctx.channel.id

    if content is None:
        print(
            "Warning: Code 10. Check out https://github.com/namuKR/d4-buttons/wiki#code-10-1 for more info.")
        content = '​'  # zero width space inside it!
    else:
        cntt = content


def send(*_buttons):
    global chn, cntt
    i = 0
    r = []
    for button in _buttons:
        i += 1
        if i > 5:
            raise SyntaxError.msg == "You can only request 5 buttons at a time."
        self = button
        headers = {"Authorization": f"Bot {Token}"}
        headers["Content-Type"] = "application/json"
        # channel_id: str
        # if ctx is None and id is None:
        #     raise TypeError.msg == "Error: Code 10. Check out https://github.com/namuKR/d4-buttons/wiki#code-10 for more info."
        # elif ctx is not None and id is not None:
        #     channel_id = ctx.channel.id
        # elif ctx is None and id is not None:
        #     channel_id = id
        # elif ctx is not None and id is None:
        #     channel_id = ctx.channel.id

        # if content is None:
        #     print(
        #         "Warning: Code 10. Check out https://github.com/namuKR/d4-buttons/wiki#code-10-1 for more info.")
        #     content = '​'  # zero width space inside it!

        if self.url is not None and self.customId is None and self.emoji is None:  # Link Button

            r.append({
                "type": 2,
                "label": self.text,
                "style": 5,
                "url": self.url,
                "disabled": self.disabled if self.disabled else False
            })

        elif self.url is not None and self.customId is None and self.emoji is not None:

            r.append({
                "type": 2,
                "label": self.text,
                "style": 5,
                "url": self.url,
                "disabled": self.disabled if self.disabled else False,
                "emoji": self.emoji
            })

        elif self.emoji is not None:
            if self.customId is None:
                raise TypeError.msg == "Buttons that are not Link style must have customId values."
            elif self.style is None or self.text is None:
                raise TypeError.msg == "Button.style or Button.text missing"

            r.append({
                "type": 2,
                "label": self.text,
                "style": self.style,
                "emoji": self.emoji,
                "custom_id": self.customId,
                "disabled": self.disabled if self.disabled else False
            })

        elif self.url is not None and self.customId is not None:
            raise TypeError.msg == "Link Buttons should not have customId values."
        elif self.url is None and self.customId is None:
            raise TypeError.msg == "Buttons that are not Link style must have customId values."
        elif self.style is None or self.text is None:
            raise TypeError.msg == "Button.style or Button.text missing"
        elif type(self.style) is not int:
            raise TypeError.msg == "Button.style is not int - button style must be int type."
        else:

            r.append({
                "type": 2,
                "label": self.text,
                "style": self.style,
                "custom_id": self.customId,
                "disabled": self.disabled if self.disabled else False
            })

    rq = []
    for v in range(i):
        rq.append(r[v])

    req_body: dict = {
        "content": cntt,
        "tts": False,
        "components": [
            {
                "type": 1,
                "components": [

                ]
            }
        ]

    }

    req_body["components"][0]["components"] = rq

    response = requests.post(
        v9_url + f"/channels/{chn}/messages", headers=headers, json=req_body)
    return response



# class User():
#     def __init__(self, **kwargs):
#         self.username = kwargs.get('username')
#         self.public_flags = kwargs.get('public_flags')
#         self.id = kwargs.get('id')
#         self.discriminator = kwargs.get('discriminator')
#         self.avatar = kwargs.get('avatar')
#         self.roles = kwargs.get('roles')
#         self.premium_since = kwargs.get('premium_since')
#         self.pending = bool(kwargs.get('pending'))
#         self.nick = kwargs.get('nick')
#         self.mute = bool(kwargs.get('mute'))
#         self.joined_at = kwargs.get('joined_at')
#         self.is_pending = bool(kwargs.get('is_pending'))
#         self.hoisted_role = kwargs.get('hoisted_role')
#         self.deaf = bool(kwargs.get('deaf'))
#         self.avatar
from ._requests import *


class Member:
    def __init__(self, **kwargs):
        self.user = User(**kwargs.get('user')) if kwargs.get('user') else None
        self.nick = kwargs.get('nick') or None
        self.roles = list(kwargs.get('roles')) if kwargs.get('roles') else []
        self.joined_at = kwargs.get('joined_at')


class User:
    def __init__(self, **kwargs):
        self.username = kwargs.get('username')
        self.public_flags = int(kwargs.get('public_flags')) if kwargs.get(
            'public_flags') else None
        self.id = kwargs.get('id')
        self.discriminator = kwargs.get('discriminator')
        self.avatar = kwargs.get('avatar')
        self.bot = bool(kwargs.get('bot')) if kwargs.get('bot') else False
        self.system = bool(kwargs.get('system')) if kwargs.get(
            'system') else False
        self.mfa_enabled = bool(kwargs.get('mfa_enabled')) if kwargs.get(
            'mfa_enabled') else False
        self.locale = kwargs.get('locale') or None
        self.verified = bool(kwargs.get('verified')) if kwargs.get(
            'verified') else False
        self.email = kwargs.get('email') or None
        self.flags = int(kwargs.get('flags')) if kwargs.get('flags') else None
        self.premium_type = int(kwargs.get('premium_type')) if kwargs.get(
            'premium_type') else None
        self.webhook_id = kwargs.get('webhook_id') or None


class Connection:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.type = kwargs.get('type')
        self.revoked = bool(kwargs.get('revoked')) if kwargs.get(
            'revoked') else False
        self.integrations = list(kwargs.get('integrations')) if kwargs.get(
            'integrations') else None
        self.verified = bool(kwargs.get('verified'))
        self.friend_sync = bool(kwargs.get('friend_sync'))
        self.show_activity = bool(kwargs.get('show_activity'))
        self.visibility = int(kwargs.get('visibility'))


class Overwrite:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.type = kwargs.get('type')
        self.allow = kwargs.get('allow')
        self.deny = kwargs.get('deny')


class ThreadMetadata:
    def __init__(self, **kwargs):
        self.archived = bool(kwargs.get('archived'))
        self.auto_archive_duration = kwargs.get('auto_archive_duration')
        self.archive_timestamp = kwargs.get('archive_timestamp')
        self.locked = kwargs.get('locked') or False


class ThreadMember:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id') or None
        self.user_id = kwargs.get('user_id') or None
        self.join_timestamp = kwargs.get('join_timestamp') or None
        self.flags = kwargs.get('flags') or None


class Client_Status:
    def __init__(self, **kwargs):
        self.desktop = kwargs.get('desktop') or None
        self.mobile = kwargs.get('mobile') or None
        self.web = kwargs.get('web') or None


class Activity:
    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        self.type = int(kwargs.get('type'))
        self.url = kwargs.get('url') if self.type == 1 else None
        self.created_at = int(kwargs.get('created_at'))
        self.timestamps = ActivityTimestamps(
            **kwargs.get('timestamps')) if kwargs.get('timestamps') else None
        self.application_id = int(kwargs.get('application_id'))
        self.details = kwargs.get('details') or None
        self.state = kwargs.get('state') or None
        self.emoji = ActivityEmoji(
            **kwargs.get('emoji')) if kwargs.get('emoji') else None
        self.party = ActivityParty(
            **kwargs.get('party')) if kwargs.get('party') else None
        self.assets = ActivityAssets(
            **kwargs.get('assets')) if kwargs.get('assets') else None
        self.secrets = ActivitySecrets(
            **kwargs.get('secrets')) if kwargs.get('secrets') else None
        self.instance = bool(kwargs.get('instance') or None)
        self.flags = kwargs.get('flags') or None
        self.buttons = list(ActivityButtons(
            **kwargs.get('buttons'))) if kwargs.get('buttons') else None


class ActivityButtons:
    def __init__(self, **kwargs):
        self.label = kwargs.get('label')
        self.url = kwargs.get('url')


class ActivityAssets:
    def __init__(self, **kwargs):
        self.large_image = kwargs.get('large_image') or None
        self.large_text = kwargs.get('large_text') or None
        self.small_image = kwargs.get('small_image') or None
        self.small_text = kwargs.get('small_text') or None


class ActivityFlags:
    INSTANCE = 1 << 0
    JOIN = 1 << 1
    SPECTATE = 1 << 2
    JOIN_REQUEST = 1 << 3
    SYNC = 1 << 4
    PLAY = 1 << 5


class ActivitySecrets:
    def __init__(self, **kwargs):
        self.join = kwargs.get('join') or None
        self.spectate = kwargs.get('spectate') or None
        self.match = kwargs.get('match') or None


class ActivityParty:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id') or None
        self.size = list(kwargs.get('size')) if kwargs.get('size') else None


class ActivityEmoji:
    def __init__(self, **kwargs):
        self.name = str(kwargs.get('name'))
        self.id = int(kwargs.get('id')) if kwargs.get('id') else None
        self.animated = bool(kwargs.get('animated')) if kwargs.get(
            'animated') else None


class ActivityTimestamps:
    def __init__(self, **kwargs):
        self.start = int(kwargs.get('start')) if kwargs.get('start') else None
        self.end = int(kwargs.get('end')) if kwargs.get('end') else None


class Presence:
    def __init__(self, **kwargs):
        self.user = User(**kwargs.get('user'))
        self.guild_id = int(kwargs.get('guild'))
        self.status = str(kwargs.get('status'))
        self.activites = list(User(**kwargs.get('activites'))
                              if kwargs.get('activites') else []) or []
        self.client_status = Client_Status(
            **kwargs.get('client_status')) if kwargs.get('client_status') else None


class WelcomeScreenChannel:
    def __init__(self, **kwargs):
        self.channel_id = kwargs.get('channel_id')
        self.description = kwargs.get('description')
        self.emoji_id = kwargs.get('emoji_id')
        self.emoji_name = kwargs.get('emoji_name')


class WelcomeScreen:
    def __init__(self, **kwargs):
        self.description = kwargs.get('description') or None
        self.welcome_channels = list(WelcomeScreenChannel(
            **kwargs.get('welcome_channels')) if kwargs.get('welcome_channels') else []) or None


class StageInstance:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.guild_id = kwargs.get('guild_id')
        self.channel_id = kwargs.get('channel_id')
        self.topic = kwargs.get('topic')
        self.privacy_level = kwargs.get('privacy_level')
        self.discoverable_disabled = kwargs.get('discoverable_disabled')


class Sticker:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.pack_id = kwargs.get('pack_id')
        self.name = kwargs.get('name')
        self.description = kwargs.get('description')
        self.tag = kwargs.get('tag')
        self.type = kwargs.get('type')
        self.format_type = kwargs.get('format_type')
        self.available = kwargs.get('available') or None
        self.guild_id = kwargs.get('guild_id') or None
        self.user = User(**kwargs.get('user')) if kwargs.get('user') else None
        self.sort_value = kwargs.get('sort_value') or None


class RoleTags:
    def __init__(self, **kwargs):
        self.bot_id = kwargs.get('bot_id')
        self.integration_id = kwargs.get('integration_id')
        self.premium_subscriber = kwargs.get('premium_subscriber') or None


class Role:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.color = kwargs.get('color')
        self.hoist = kwargs.get('hoist')
        self.position = kwargs.get('position')
        self.premissions = kwargs.get('premissions')
        self.managed = kwargs.get('managed')
        self.mentionable = kwargs.get('mentionable')
        self.tags = RoleTags(**kwargs.get('tags')
                             ) if kwargs.get('tags') else None


class Guild:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.icon = kwargs.get('icon') or None
        self.icon_hash = kwargs.get('icon_hash') or None
        self.splash = kwargs.get('splash') or None
        self.discovery_splash = kwargs.get('discovery_splash') or None
        self.owner = kwargs.get('owner') or None
        self.owner_id = kwargs.get('owner_id') or None,
        self.permissions = kwargs.get('permissions') or None,
        self.afk_channel_id = kwargs.get('afk_channel_id') or None,
        self.afk_timeout = kwargs.get('afk_timeout') or None,
        self.widget_enabled = bool(kwargs.get('widget_enabled')) if kwargs.get(
            'widget_enabled') else None
        self.widget_channel_id = kwargs.get('widget_channel_id') or None
        self.verification_level = int(kwargs.get('verification_level'))
        self.default_message_notifications = int(
            kwargs.get('default_message_notifications'))
        self.explicit_content_filter = int(
            kwargs.get('explicit_content_filter'))
        self.roles = list(kwargs.get('roles')) if kwargs.get('roles') else []
        self.emojis = list(kwargs.get('emojis')
                           ) if kwargs.get('emojis') else []
        self.features = list(kwargs.get('features')
                             ) if kwargs.get('features') else []
        self.mfa_level = int(kwargs.get('mfa_level'))
        self.application_id = int(kwargs.get('application_id')) if kwargs.get(
            'application_id') else None
        self.system_channel_id = int(kwargs.get('system_channel_id')) if kwargs.get(
            'system_channel_id') else None
        self.system_channel_flags = int(kwargs.get('system_channel_flags'))
        self.rules_channel_id = int(kwargs.get('rules_channel_id')) if kwargs.get(
            'rules_channel_id') else None
        self.joined_at = kwargs.get('joined_at') or None
        self.large = bool(kwargs.get('large')) if kwargs.get(
            'large') else False
        self.unavailable = bool(kwargs.get('unavailable')) if kwargs.get(
            'unavailable') else False
        self.member_count = int(kwargs.get('member_count')) if kwargs.get(
            'member_count') else 0
        self.voice_states = list(kwargs.get('voice_states')) if kwargs.get(
            'voice_states') else []
        self.members = list(
            Member(**i) for i in kwargs.get('members')) if kwargs.get('members') else []
        self.channels = list(Channel(**i)
                             for i in kwargs.get('channels')) if kwargs.get('channels') else []
        self.threads = list(
            Channel(**i) for i in kwargs.get('threads')) if kwargs.get('threads') else []
        self.presences = list(Presence(**kwargs.get('presences'))
                              if kwargs.get('presences') else []) if kwargs.get('presences') else []
        self.max_presences = kwargs.get('max_presences')
        self.max_members = kwargs.get('max_members')
        self.vanity_url_code = kwargs.get('vanity_url_code')
        self.description = kwargs.get('description')
        self.banner = kwargs.get('banner')
        self.premium_tier = kwargs.get('premium_tier')
        self.premium_subscription_count = kwargs.get(
            'premium_subscription_count') or None
        self.preferred_locale = kwargs.get('preferred_locale')
        self.public_updates_channel_id = kwargs.get(
            'public_updates_channel_id')
        self.max_video_channel_users = kwargs.get(
            'max_video_channel_users') or None
        self.approximate_member_count = kwargs.get(
            'approximate_member_count') or None
        self.approximate_presence_count = kwargs.get(
            'approximate_presence_count') or None
        self.welcome_screen = WelcomeScreen(
            **kwargs.get('welcome_screen')) if kwargs.get('welcome_screen') else None
        self.nsfw_level = kwargs.get('nsfw_level')
        self.stage_instances = list(StageInstance(
            **kwargs.get('stage_instances')) if kwargs.get('stage_instances') else []) or []
        self.stickers = list(Sticker(**kwargs.get('stickers'))
                             if kwargs.get('stickers') else []) or None


class ChannelMention:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.guild_id = kwargs.get('guild_id')
        self.type = kwargs.get('type')
        self.name = kwargs.get('name')


class Attachment:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.filename = kwargs.get('filename')
        self.content_type = kwargs.get('content_type') or None
        self.size = kwargs.get('size')
        self.url = kwargs.get('url')
        self.proxy_url = kwargs.get('proxy_url')
        self.height = kwargs.get('height') or None
        self.width = kwargs.get('width') or None


class Embed:
    def __init__(self, **kwargs):
        self.dict_: dict
        self.title = kwargs.get('title') or None
        self.dict_["title"] = self.title
        self.type = kwargs.get('type') or None
        self.dict_["type"] = self.type
        self.description = kwargs.get('description') or None
        self.dict_["description"] = self.description
        self.url = kwargs.get('url') or None
        self.dict_["url"] = self.url
        self.timestamp = kwargs.get('timestamp') or None
        self.dict_["timestamp"] = self.timestamp
        self.color = kwargs.get('color') or None
        self.dict_["color"] = self.color
        self.footer = EmbedFooter(
            **kwargs.get('footer')) if kwargs.get('footer') else None
        self.dict_["footer"] = self.footer
        self.image = EmbedImage(**kwargs.get('image')
                                ) if kwargs.get('image') else None
        self.dict_["image"] = self.image
        self.thumbnail = EmbedThumbnail(
            **kwargs.get('thumbnail')) if kwargs.get('thumbnail') else None
        self.dict_["thumbnail"] = self.thumbnail
        self.video = EmbedVideo(**kwargs.get('video')
                                ) if kwargs.get('video') else None
        self.dict_["video"] = self.video
        self.provider = EmbedProvider(
            **kwargs.get('provider')) if kwargs.get('provider') else None
        self.dict_["provider"] = self.provider
        self.author = EmbedAuthor(
            **kwargs.get('author')) if kwargs.get('author') else None
        self.dict_["author"] = self.author
        self.fields = EmbedField(
            **kwargs.get('fields')) if kwargs.get('fields') else None
        self.dict_["fields"] = self.fields

    def json(self):
        return_dict: dict = {}
        for name, value in self.dict_:
            if value is not None:
                return_dict[name] = value

        return return_dict


class EmbedThumbnail:
    def __init__(self, **kwargs):
        self.url = kwargs.get('url') or None
        self.proxy_url = kwargs.get('proxy_url') or None
        self.height = kwargs.get('height') or None
        self.width = kwargs.get('width') or None


class EmbedVideo:
    def __init__(self, **kwargs):
        self.url = kwargs.get('url') or None
        self.proxy_url = kwargs.get('proxy_url') or None
        self.height = kwargs.get('height') or None
        self.width = kwargs.get('width') or None


class EmbedImage:
    def __init__(self, **kwargs):
        self.url = kwargs.get('url') or None
        self.proxy_url = kwargs.get('proxy_url') or None
        self.height = kwargs.get('height') or None
        self.width = kwargs.get('width') or None


class EmbedProvider:
    def __init__(self, **kwargs):
        self.name = kwargs.get('name') or None
        self.url = kwargs.get('url') or None


class EmbedAuthor:
    def __init__(self, **kwargs):
        self.name = kwargs.get('name') or None
        self.url = kwargs.get('url') or None
        self.icon_url = kwargs.get('icon_url') or None
        self.proxy_icon_url = kwargs.get('proxy_icon_url') or None


class EmbedFooter:
    def __init__(self, **kwargs):
        self.text = kwargs.get('text')
        self.icon_rl = kwargs.get('icon_url') or None
        self.proxy_icon_url = kwargs.get('proxy_icon_url') or None


class EmbedField:
    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        self.value = kwargs.get('value')
        self.inline = kwargs.get('inline') or None


class Emoji:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id') or None
        self.name = kwargs.get('name') or None
        self.roles = list(Role(**kwargs.get('roles'))
                          if kwargs.get('roles') else []) or None
        self.user = User(**kwargs.get('user')) if kwargs.get('user') else None
        self.require_colons = kwargs.get('require_colons') or None
        self.managed = kwargs.get('managed') or None
        self.animated = kwargs.get('animated') or None
        self.available = kwargs.get('available') or None


class Reaction:
    def __init__(self, **kwargs):
        self.count = kwargs.get('count')
        self.me = bool(kwargs.get('me'))
        self.emoji = Emoji(**kwargs.get('emoji')
                           ) if kwargs.get('emoji') else None


class MessageActivity:
    def __init__(self, **kwargs):
        self.type = kwargs.get('type')
        self.party_id = kwargs.get('party_id') or None


class Application:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.icon = kwargs.get('icon') or None
        self.description = kwargs.get('description')
        self.rpc_origins = list((i for i in kwargs.get(
            'rpc_origins')) if kwargs.get('rpc_origins') else []) or False
        self.bot_public = bool(kwargs.get('bot_public'))
        self.bot_require_code_grant = bool(
            kwargs.get('bot_require_code_grant'))
        self.terms_of_service_url = kwargs.get('terms_of_service_url') or None
        self.privacy_policy_url = kwargs.get('privacy_policy_url') or None
        self.owner = User(**kwargs.get('owner')
                          ) if kwargs.get('owner') else None
        self.summary = kwargs.get('summary')
        self.verify_key = kwargs.get('verify_key')
        self.team = Team(**kwargs.get('team')) if kwargs.get('team') else None
        self.guild_id = kwargs.get('guild_id') or None
        self.primary_sku_id = kwargs.get('primary_sku_id') or None
        self.slug = kwargs.get('slug') or None
        self.cover_image = kwargs.get('cover_image') or None
        self.flags = kwargs.get('flags') or None


class Team:
    def __init__(self, **kwargs):
        self.icon = kwargs.get('icon') or None
        self.id = kwargs.get('id') or None
        self.members = list(TeamMember(**kwargs.get('members'))
                            ) if kwargs.get('members') else None
        self.name = kwargs.get('name')
        self.owner_user_id = kwargs.get('owner_user_id')


class TeamMember:
    def __init__(self, **kwargs):
        self.membership_state = kwargs.get('membership_state')
        self.permissions = list(kwargs.get('permissions'))
        self.team_id = kwargs.get('team_id')
        self.user = User(**kwargs.get('user')) if kwargs.get('user') else None


class MessageReference:
    def __init__(self, **kwargs):
        self.message_id = kwargs.get('message_id')
        self.channel_id = kwargs.get('channel_id') or None
        self.guild_id = kwargs.get('guild_id') or None
        self.fail_if_not_exists = kwargs.get('fail_if_not_exists') or None


class SelectMenu:
    def __init__(self, **kwargs):
        self.type = kwargs.get('type')
        self.custom_id = kwargs.get('custom_id')
        self.options = list(SelectOption(**kwargs.get('options'))
                            if kwargs.get('options') else []) or None
        self.placeholder = kwargs.get('placeholder') or None
        self.min_values = kwargs.get('min_values') or None
        self.max_values = kwargs.get('max_values') or None
        self.disabled = kwargs.get('disabled') or None


class SelectOption:
    def __init__(self, **kwargs):
        self.label = kwargs.get('label')
        self.value = kwargs.get('value')
        self.description = kwargs.get('description') or None
        self.emoji = Emoji(**kwargs.get('emoji')
                           ) if kwargs.get('emoji') else None
        self.default = kwargs.get('default') or None


class AllowedMention:
    def __init__(self, **kwargs):
        self.parse = list(AllowedMentionType(**i)
                          for i in kwargs.get('parse')) if kwargs.get('parse') else None
        self.roles = kwargs.get('roles')
        self.users = kwargs.get('users')
        self.replied_user = kwargs.get('replied_user') or False


class Component:
    def __init__(self, **kwargs):
        self.type = kwargs.get('type')
        self.custom_id = kwargs.get('custom_id') or None
        self.disabled = kwargs.get('disabled') or None
        self.style = kwargs.get('style') or None
        self.label = kwargs.get('label') or None
        self.emoji = Emoji(**kwargs.get('emoji')
                           ) if kwargs.get('emoji') else None
        self.url = kwargs.get('url') or None
        self.options = list(SelectOption(
            **i) for i in kwargs.get('options')) if kwargs.get('options') else None
        self.placeholder = kwargs.get('placeholder') or None
        self.min_values = kwargs.get('min_values') or None
        self.max_values = kwargs.get('max_values') or None
        self.components = list(Component(
            **i) for i in kwargs.get('components')) if kwargs.get('components') else None


class Interaction:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.application_id = kwargs.get('application_id')
        self.type = kwargs.get('type')
        self.data = InteractionData(**kwargs.get(
            'data')) if kwargs.get('data') else None
        self.guild_id = kwargs.get('guild_id') or None
        self.channel_id = kwargs.get('channel_id') or None
        self.member = Member(**kwargs.get('member')
                             ) if kwargs.get('member') else None
        self.user = User(**kwargs.get('user')) if kwargs.get('user') else None
        self.token = kwargs.get('token')
        self.version = kwargs.get('version')
        self.message = Message(**kwargs.get('message')
                               ) if kwargs.get('message') else None

    async def reply(self, msg):
        if isinstance(msg, Message):
            pass
        req_body = {
            'type': 4,
            'data': {
                'content': msg
            }
        }
        post(
            f'/interactions/{self.id}/{self.token}/callback', req_body)

    async def reply_raw(self, message, type: int = 4, embeds: list = None, tts=None, file=None, payload_json=None, allowed_mentions: AllowedMention = None, message_reference: MessageReference = None, components: Component = None, sticker_ids: list = None):
        if embeds:
            for i in embeds:
                if not isinstance(i, Embed):
                    raise TypeError(
                        "embeds must be an instance of <class 'Embed'>")
        base_json = {
            "content": message,
        }
        if embeds:
            base_json["embeds"] = embeds
        if tts:
            base_json["tts"] = tts

        if file:
            base_json["file"] = file
        if payload_json:
            base_json["payload_json"] = payload_json
        if allowed_mentions:
            base_json["allowed_mentions"] = allowed_mentions
        if message_reference:
            base_json["message_reference"] = message_reference
        if components:
            base_json["components"] = components
        if sticker_ids:
            base_json["sticker_ids"] = sticker_ids

        req_body = {
            "type": type,
            "data": base_json,
        }

        post(
            f'/interactions/{self.id}/{self.token}/callback', req_body)


class InteractionData:  # https://discord.com/developers/docs/interactions/receiving-and-responding#interaction-object-interaction-type
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.type = kwargs.get('type')
        self.resolved = ResolvedData(
            **kwargs.get('resolved')) if kwargs.get('resolved') else None
        # TODO add when support application command (slash commands)
        # self.options
        self.custom_id = kwargs.get('custom_id')
        self.component_type = kwargs.get('component_type') or None
        self.values = list(SelectOption(
            **i) for i in kwargs.get('values')) if kwargs.get('values') else None
        self.target_id = kwargs.get('target_id') or None


class ResolvedData:
    def __init__(self, **kwargs):
        self.users = map(User(**kwargs.get('users')),
                         User.objects.filter()) if kwargs.get('users') else None
        self.members = map(Member(**kwargs.get('members')),
                           Member.objects.filter()) if kwargs.get('members') else None
        self.roles = map(Role(**kwargs.get('roles')),
                         Role.objects.filter()) if kwargs.get('roles') else None
        self.channels = map(Channel(**kwargs.get('channels')),
                            Channel.objects.filter()) if kwargs.get('channels') else None
        self.messages = map(Message(**kwargs.get('messages')),
                            Message.objects.filter()) if kwargs.get('messages') else None


class MessageInteraction:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.type = Interaction(**kwargs.get('type'))
        self.name = kwargs.get('name')
        self.user = User(**kwargs.get('user')) if kwargs.get('user') else None


class StickerItem:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.format_type = kwargs.get('format_type')


class Message:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.channel_id = kwargs.get('channel_id')
        self.guild_id = kwargs.get('guild_id')
        self.author = User(**kwargs.get('author'))
        self.member = (Member(**i) for i in kwargs.get('member')
                       ) if kwargs.get('member') else None
        self.content = kwargs.get('content')
        self.timestamp = kwargs.get('timestamp')
        self.edited_timestamp = kwargs.get('edited_timestamp') or None
        self.tts = bool(kwargs.get('tts'))
        self.mention_everyone = bool(kwargs.get('mention_everyone'))
        self.mentions = list((User(**i) for i in kwargs.get('mentions'))
                             if kwargs.get('mentions') else []) or None
        self.mention_roles = list(Role(
            **kwargs.get("mention_roles")) if kwargs.get("mention_roles") else []) or None
        self.mention_channels = ChannelMention(
            **kwargs.get('mention_channels')) if kwargs.get('mention_channels') else None
        self.attachments = list(Attachment(
            **kwargs.get('attachments')) if kwargs.get('attachments') else []) or None
        self.embeds = list(Embed(**kwargs.get('embeds'))
                           if kwargs.get('embeds') else []) or None
        self.reactions = list(Reaction(**kwargs.get('reactions'))
                              if kwargs.get('reactions') else []) or None
        self.nonce = kwargs.get('nonce') or None
        self.pinned = bool(kwargs.get('pinned'))
        self.webhook_id = None if not kwargs.get(
            'webhook_id') else kwargs.get('webhook_id')
        self.type = kwargs.get('type')
        self.activity = MessageActivity(
            **kwargs.get('activity')) if kwargs.get('activity') else None
        self.application = Application(
            **kwargs.get('application')) if kwargs.get('application') else None
        self.application_id = kwargs.get('application_id') or None
        self.message_reference = MessageReference(
            **kwargs.get('message_reference')) if kwargs.get('message_reference') else None
        self.flags = kwargs.get('flags') or None
        self.referenced_message = Message(
            **kwargs.get('referenced_message')) if kwargs.get('referenced_message') else None
        self.interaction = MessageInteraction(
            **kwargs.get('interaction')) if kwargs.get('interaction') else None
        self.thread = Channel(**kwargs.get('thread')
                              ) if kwargs.get('thread') else None
        self.components = list(Component(
            **i) for i in kwargs.get('components')) if kwargs.get('components') else None
        self.sticker_items = list(StickerItem(
            **i) for i in kwargs.get('sticker_items')) if kwargs.get('sticker_items') else None
        self.stickers = list(Sticker(
            **i) for i in kwargs.get('stickers')) if kwargs.get('stickers') else None
        self.channel = self.get_channel()

    def get_channel(self):
        return Channel(
            **get(f'/channels/{self.channel_id}').json()) if self.channel_id else None


class AllowedMentionType:
    def __init__(self, type_):
        self.type = "roles" if type_ == "roles" else (
            "users" if type_ == "users" else ("everyone" if type_ == "everyone" else None))


class Channel:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.type = kwargs.get('type') or 1
        self.guild_id = kwargs.get('guild_id') or None
        self.position = int(kwargs.get('position')) if kwargs.get(
            'position') else None
        self.permission_overwrites = list(Overwrite(**i) for i in kwargs.get(
            'permission_overwrites')) if kwargs.get('permission_overwrites') else None
        self.name = kwargs.get('name') or None
        self.topic = kwargs.get('topic') or None
        self.nsfw = kwargs.get('nsfw') or None
        self.last_message_id = kwargs.get('last_message_id') or None
        self.birate = kwargs.get('birate') or None
        self.user_limit = int(kwargs.get('user_limit')
                              ) if kwargs.get('user_limit') else None
        self.rate_limit_per_user = kwargs.get('rate_limit_per_user') or None
        self.recipients = list(
            User(**i) for i in kwargs.get('recipients')) if kwargs.get('recipients') else []
        self.icon = kwargs.get('icon') or None
        self.owner_id = kwargs.get('owner_id') or None
        self.application_id = kwargs.get('application_id') or None
        self.parent_id = kwargs.get('parent_id') or None
        self.last_pin_timestamp = kwargs.get('last_pin_timestamp') or None
        self.rtc_region = kwargs.get('rtc_region') or None
        self.video_quality_mode = kwargs.get('video_quality_mode') or None
        self.message_count = kwargs.get('message_count') or None
        self.member_count = kwargs.get('member_count') or None
        self.thread_metadata = ThreadMetadata(
            **kwargs.get('thread_metadata')) if kwargs.get('thread_metadata') else None
        self.member = ThreadMember(
            **kwargs.get('member')) if kwargs.get('member') else None
        self.default_auto_archive_duration = kwargs.get(
            'default_auto_archive_duration') or None
        self.permissions = kwargs.get('permissions') or None

    async def send(self, message, channelId=None, embeds: list = None, tts=None, file=None, payload_json=None, allowed_mentions: AllowedMention = None, message_reference: MessageReference = None, components: Component = None, sticker_ids: list = None):
        if embeds:
            for i in embeds:
                if not isinstance(i, Embed):
                    raise TypeError(
                        "embeds must be an instance of <class 'Embed'>")
        base_json = {
            "content": message,
        }
        if embeds:
            base_json["embeds"] = embeds
        if tts:
            base_json["tts"] = tts

        if file:
            base_json["file"] = file
        if payload_json:
            base_json["payload_json"] = payload_json
        if allowed_mentions:
            base_json["allowed_mentions"] = allowed_mentions
        if message_reference:
            base_json["message_reference"] = message_reference
        if components:
            base_json["components"] = components
        if sticker_ids:
            base_json["sticker_ids"] = sticker_ids
        post(
            f"/channels/{channelId if channelId else self.id}/messages", base_json)



from ._exception import *
from ._interactions import *
command_dict: dict = {}


class Command:
    def __init__(self, func=None, *args, **kwargs):
        self.func = func
        self.name = kwargs.get('name'),
        self.aliases = kwargs.get('aliases'),
        self.slash = kwargs.get('slash'),
        self.prefix = kwargs.get('prefix'),

    def add_command(self, fx):
        global Log
        if not isinstance(fx, Command):
            raise TypeError("Not instance of <class 'command'>")

        if isinstance(self, Command):
            fx.parent = self

        if fx.name[0] in command_dict:
            raise CommandRegistrationError().error()

        command_dict[fx.name[0]] = fx.func
        for alias in fx.aliases[0]:
            if alias in command_dict.keys():
                self.remove_command(fx.name[0])
                raise CommandRegistrationError().error()
            command_dict[alias] = fx.func
        Log("Command Successfully registered")
        return "Command Successfully Registered"

    def remove_command(self, name):
        cmd = command_dict.pop(name, None)

        if cmd is None:
            return None

        if name in cmd.aliases:
            return cmd

        for alias in cmd.aliases:
            __cmd__ = command_dict.pop(alias, None)
            if __cmd__ not in (None, cmd):
                command_dict[alias] = __cmd__
        return cmd


### EVENTS ###
event_dict = {}



import setuptools

with open("README", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="d4-py",
    version="0.0.0.10a3",
    author="ast.08",
    author_email="<ast.08e@gmail.com>",
    description="d4.py pre-alpha v0.0.0.10a",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/namuKR/d4.py",
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
    ],
    python_requires='>=3.6',
    install_requires=['requests', 'asyncio', 'aiohttp']
)
